/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 * 	This product includes software developed by the Network Research
 * 	Group at Lawrence Berkeley National Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/*
 * a "flow manager".  This is an object used to keep track of
 * various statistics associated with per-flow information.
 */

#ifndef lint
static char rcsid[] =
    "@(#) $Header: flow.cc,v 1.32 97/02/06 11:25:55 floyd Exp $ (LBL)";
#endif

#include <stdlib.h>
#include "flow.h"
#include "link.h"

// tcl linkage
class FlowManagerMatcher : public Matcher {
    public:
        inline FlowManagerMatcher() : Matcher("flowmgr") { }
        TclObject* match(const char* id) {
		if (id == 0) {
			printf("flowmanager flavor not specified\n");
			return 0;
		}
		/* what packet fields are used to differentiate a 'flow' */
                if (strcmp(id, "class") == 0)
                        return (new ClassFlowManager);
                else if (strcmp(id, "src") == 0)
                        return (new SrcFlowManager);
                else if (strcmp(id, "dst") == 0)
                        return (new DstFlowManager);
                else if (strcmp(id, "srcdst") == 0)
                        return (new SrcDstFlowManager);
                else if (strcmp(id, "dstsrc") == 0)
                        return (new SrcDstFlowManager);
                else if (strcmp(id, "srcdstcls") == 0)
                        return (new SrcDstClassFlowManager);
                else if (strcmp(id, "redsrcdstcls") == 0)
                        return (new SrcDstClassFlowManager(FTYPE_RED));
                return (0);
        }
} flowmgrmatcher;


#ifndef MAXINT
#define	MAXINT	2147483647
#endif

double FlowManager::computeFrac(int first, int second)
{
	if (second == 0) {
		return (0.0);
	} else {
		return ((double) first/second);
	}
}

/*
 * tcl "member functions" on a flowmgr object
 */
int FlowManager::command(int argc, const char*const* argv)
{
        Tcl& tcl = Tcl::instance();
        if (argc == 2) {
		// remove all flows
		if (strcmp(argv[1], "freeflows") == 0) {
			freeflows();
			return (TCL_OK);
		}
		// per-flow trace file support
                if (strcmp(argv[1], "detach") == 0) {
                        channel_ = 0;
                        return (TCL_OK);
                }
		// flush output to trace-file
                if (strcmp(argv[1], "flush") == 0) {
                        Tcl_Flush(channel_);
                        return (TCL_OK);
                }
		// resets all counters in flows
                if (strcmp(argv[1], "resetcounters") == 0) {
                        resetcounters();
                        return (TCL_OK);
                }
		// dumps all flows/categories...
		//	[also see "dump $categ" below]
                if (strcmp(argv[1], "dump") == 0) {
                        return(dumpflows());
                }
		// returns number of flows
                if (strcmp(argv[1], "flows") == 0) {
			char *bp = tcl.buffer();
			sprintf(bp, "%u", nflows_);
			tcl.result(bp);
			return (TCL_OK);
                }
		// returns number of flows with drops
                if (strcmp(argv[1], "fwdrops") == 0) {
			char *bp = tcl.buffer();
			sprintf(bp, "%u", fwdrops_);
			tcl.result(bp);
			return (TCL_OK);
                }
		// returns number of flows with arrivals
                if (strcmp(argv[1], "fwarrivals") == 0) {
			char *bp = tcl.buffer();
			sprintf(bp, "%u", fwarrivals_);
			tcl.result(bp);
			return (TCL_OK);
                }
		// returns number of total pkt arrivals
                if (strcmp(argv[1], "narrivals") == 0) {
			char *bp = tcl.buffer();
			sprintf(bp, "%u %u",
				allarrivals_.pkt_arrivals_,
				allarrivals_.byte_arrivals_);
			tcl.result(bp);
			return (TCL_OK);
                }
		// returns number of total pkt drops
                if (strcmp(argv[1], "ndrops") == 0) {
			char *bp = tcl.buffer();
			sprintf(bp, "%u %u",
				alldrops_.pkt_drops_,
				alldrops_.byte_drops_);
			tcl.result(bp);
			return (TCL_OK);
                }
		// return the flow with the highest drop metric 
		if (strcmp(argv[1], "maxmetric") == 0) {
			register i = 0, j = 0;
			double max = -1.0;
			double metric, unforcedFraction, forcedFraction;
			double unforcedMetric, forcedMetric;
			register Flow *fp = 0;
			char *bp = tcl.buffer();
			for (i = 0; i < nflows_; i++) {
				metric = 0.0;
				// forced drops, category 0:
				forcedMetric = computeFrac(
				  flows_[i]->bdrops(0),
				  droptotals_[0].byte_drops_);
				forcedFraction = computeFrac(
				  droptotals_[0].pkt_drops_,
				  alldrops_.pkt_drops_);
				metric += forcedMetric * forcedFraction;
				// unforced drops, category 1:
				unforcedMetric = computeFrac(
				  flows_[i]->pdrops(1), 
				  droptotals_[1].pkt_drops_);
				unforcedFraction = computeFrac(
				  droptotals_[1].pkt_drops_,
				  alldrops_.pkt_drops_);
				metric += unforcedMetric * unforcedFraction;
				if (metric > max) {
					fp = flows_[i];
					max = metric;
				}
			}
			sprintf(bp, "%s %5.2f %d", fp->name(), 100*max, fp->fclass());
			tcl.result(bp);
			return (TCL_OK);
		}
		// return the flow with the lowest drop metric 
		if (strcmp(argv[1], "minmetric") == 0) {
			register i = 0, j = 0;
			double min = 20000.0;
			double metric, unforcedFraction, forcedFraction;
			double unforcedMetric, forcedMetric;
			register Flow *fp = 0;
			char *bp = tcl.buffer();
			for (i = 0; i < nflows_; i++) {
				metric = 0.0;
				// forced drops, category 0:
				forcedMetric = computeFrac(
				  flows_[i]->bdrops(0),
				  droptotals_[0].byte_drops_);
				forcedFraction = computeFrac(
				  droptotals_[0].pkt_drops_,
				  alldrops_.pkt_drops_);
				metric += forcedMetric * forcedFraction;
				// unforced drops, category 1:
				unforcedMetric = computeFrac(
				  flows_[i]->pdrops(1), 
				  droptotals_[1].pkt_drops_);
				unforcedFraction = computeFrac(
				  droptotals_[1].pkt_drops_,
				  alldrops_.pkt_drops_);
				metric += unforcedMetric * unforcedFraction;
				if (metric < min ) {
					fp = flows_[i];
					min = metric;
				}
			}

			sprintf(bp, "%s %5.2f %d", fp->name(), 100*min, fp->fclass());
			tcl.result(bp);
			return (TCL_OK);
		}
		// return the flow with the most pkt drops
		if (strcmp(argv[1], "maxpdrops") == 0) {
			register i = 0, j = 0, max = 0, drops = 0;
			register Flow *fp = 0;
			char *bp = tcl.buffer();
			for (i = 0; i < nflows_; i++) {
				drops = 0;
				for (j = 0; j < ndcat_; j++)
					drops += flows_[i]->pdrops(j);
				if (drops > max) {
					fp = flows_[i];
					max = drops;
				}
			}
			sprintf(bp, "%s %d", fp->name(), max);
			tcl.result(bp);
			return (TCL_OK);
		}
		// return the flow with the most byte drops
		if (strcmp(argv[1], "maxbdrops") == 0) {
			register i = 0, j = 0, max = 0, drops = 0;
			register Flow *fp = 0;
			char *bp = tcl.buffer();
			for (i = 0; i < nflows_; i++) {
				drops = 0;
				for (j = 0; j < ndcat_; j++)
					drops += flows_[i]->bdrops(j);
				if (drops > max) {
					fp = flows_[i];
					max = drops;
				}
			}
			sprintf(bp, "%s %d", fp->name(), max);
			tcl.result(bp);
			return (TCL_OK);
		}
		// return the flow with the minimum pkt drops
		if (strcmp(argv[1], "minpdrops") == 0) {
			register i = 0, j = 0, min = MAXINT, drops = 0;
			register Flow *fp = 0;
			char *bp = tcl.buffer();
			for (i = 0; i < nflows_; i++) {
				drops = 0;
				for (j = 0; j < ndcat_; j++)
					drops += flows_[i]->pdrops(j);
				if (drops < min) {
					fp = flows_[i];
					min = drops;
				}
			}
			sprintf(bp, "%s %d", fp->name(), min);
			tcl.result(bp);
			return (TCL_OK);
		}
		// return the flow with the minimum byte drops
		if (strcmp(argv[1], "minbdrops") == 0) {
			register i = 0, j = 0, min = MAXINT, drops = 0;
			register Flow *fp = 0;
			char *bp = tcl.buffer();
			for (i = 0; i < nflows_; i++) {
				drops = 0;
				for (j = 0; j < ndcat_; j++)
					drops += flows_[i]->bdrops(j);
				if (drops < min) {
					fp = flows_[i];
					min = drops;
				}
			}
			sprintf(bp, "%s %d", fp->name(), min);
			tcl.result(bp);
			return (TCL_OK);
		}
        } else if (argc == 3) {
		if (strcmp(argv[1], "arrivals") == 0) {
			if (strcmp(argv[2], "enable") == 0)
				enable_arrivals_default_ = 1;
			else if (strcmp(argv[2], "disable") == 0)
				enable_arrivals_default_ = 0;
			else {
				tcl.result("flowmgr: bad usage: arrivals ?");
				return (TCL_ERROR);
			}
			return (TCL_OK);
		}
		if (strcmp(argv[1], "drops") == 0) {
			if (strcmp(argv[2], "enable") == 0)
				enable_drops_default_ = 1;
			else if (strcmp(argv[2], "disable") == 0)
				enable_drops_default_ = 0;
			else {
				tcl.result("flowmgr: bad usage: drops ?");
				return (TCL_ERROR);
			}
			return (TCL_OK);
		}
				
		// attach a file for output of dump() operation
                if (strcmp(argv[1], "attach") == 0) {
                        int mode;
                        const char* id = argv[2];
                        channel_ = Tcl_GetChannel(tcl.interp(), (char*)id,
                                                  &mode);
                        if (channel_ == 0) {
                                tcl.resultf("flowmgr: can't attach %s for writing"
, id);
                                return (TCL_ERROR);
                        }
                        return (TCL_OK);
                }
		// define the number of drop categories
		if (strcmp(argv[1], "ndropcats") == 0) {
			int ncats = atoi(argv[2]);
			if (ncats <= 0 || ncats > MAXCATEGORY) {
				tcl.resultf("flowmgr: ndrop cats %d invalid [max %d]\n",
					ncats, MAXCATEGORY);
				return (TCL_ERROR);
			}
			ndcat_ = ncats;
			return (TCL_OK);
		}

		// minimum time interval for invoking pcallback
		if (strcmp(argv[1], "minimum-time") == 0) {
			double min_time = atof(argv[2]);
			if (min_time < 0.0) {
				tcl.resultf("flowmgr: invalid minimum-time: %f",
					min_time);
				return (TCL_ERROR);
			}
			alldrops_.min_time_ = min_time;
			return (TCL_OK);
		}
		// tot pkt drops theshold above which to invoke pcallback
		if (strcmp(argv[1], "total-drops-pthresh") == 0) {
			int thresh = atoi(argv[2]);
			if (thresh < 0) {
				tcl.resultf("flowmgr: invalid drop-thresh: %d",
					thresh);
				return (TCL_ERROR);
			}
			alldrops_.pkt_thresh_ = thresh;
			return (TCL_OK);
		}
		// tot bytes in drops thresh above which to invoke bcallback
		if (strcmp(argv[1], "total-drops-bthresh") == 0) {
			int thresh = atoi(argv[2]);
			if (thresh < 0) {
				tcl.resultf("flowmgr: invalid drop-thresh: %d",
					thresh);
				return (TCL_ERROR);
			}
			alldrops_.byte_thresh_ = thresh;
			return (TCL_OK);
		}
		// define the total packet drops handler
                if (strcmp(argv[1], "total-drops-pcallback") == 0) {
                        const char* cb = argv[2];
                        delete alldrops_.pcallback_;
                        if (*cb == 0)
                                alldrops_.pcallback_ = 0;
                        else {
                                alldrops_.pcallback_ = new char[strlen(cb) + 1];
                                strcpy(alldrops_.pcallback_, cb);
                        }
                        return (TCL_OK);
                }
		// define the total bytes in packet drops handler
                if (strcmp(argv[1], "total-drops-bcallback") == 0) {
                        const char* cb = argv[2];
                        delete alldrops_.bcallback_;
                        if (*cb == 0)
                                alldrops_.bcallback_ = 0;
                        else {
                                alldrops_.bcallback_ = new char[strlen(cb) + 1];
                                strcpy(alldrops_.bcallback_, cb);
                        }
                        return (TCL_OK);
                }


		// total pkt arrivals thresh above which to invoke pcallback
		if (strcmp(argv[1], "total-arrivals-pthresh") == 0) {
			int thresh = atoi(argv[2]);
			if (thresh < 0) {
				tcl.resultf("flowmgr: invalid arr-thresh: %d",
					thresh);
				return (TCL_ERROR);
			}
			allarrivals_.pkt_thresh_ = thresh;
			return (TCL_OK);
		}
		// total bytes in arrivals above which to invoke bcallback
		if (strcmp(argv[1], "total-arrivals-bthresh") == 0) {
			int thresh = atoi(argv[2]);
			if (thresh < 0) {
				tcl.resultf("flowmgr: invalid arr-thresh: %d",
					thresh);
				return (TCL_ERROR);
			}
			allarrivals_.byte_thresh_ = thresh;
			return (TCL_OK);
		}
		// define the total pkt arrivals handler
                if (strcmp(argv[1], "total-arrivals-pcallback") == 0) {
                        const char* cb = argv[2];
                        delete allarrivals_.pcallback_;
                        if (*cb == 0)
                                allarrivals_.pcallback_ = 0;
                        else {
                                allarrivals_.pcallback_ = new char[strlen(cb) + 1];
                                strcpy(allarrivals_.pcallback_, cb);
                        }
                        return (TCL_OK);
                }
		// define the total bytes in pkt arrivals handler
                if (strcmp(argv[1], "total-arrivals-bcallback") == 0) {
                        const char* cb = argv[2];
                        delete allarrivals_.bcallback_;
                        if (*cb == 0)
                                allarrivals_.bcallback_ = 0;
                        else {
                                allarrivals_.bcallback_ = new char[strlen(cb) + 1];
                                strcpy(allarrivals_.bcallback_, cb);
                        }
                        return (TCL_OK);
                }
		// usage: dump cat#
                if (strcmp(argv[1], "dump") == 0) {
			int categ = atoi(argv[2]);
			if (categ < 0 || categ > MAXCATEGORY) {
				tcl.resultf("flowmgr: invalid categ: %d",
					categ);
				return (TCL_ERROR);
			}
                        dumpflows(categ);
			return (TCL_OK);
                }

		if (strcmp(argv[1], "remflow") == 0) {
			Flow *fp = (Flow*)TclObject::lookup(argv[2]);
			if (fp == NULL) {
				tcl.resultf("can't find flow with name %s",
					argv[2]);
				return (TCL_ERROR);
			}
			remflow(fp);
			fp->setfm(NULL);
			return (TCL_OK);
		}
		if (strcmp(argv[1], "insertflow") == 0) {
			Flow *fp = (Flow*)TclObject::lookup(argv[2]);
			if (fp == NULL) {
				tcl.resultf("can't find flow with name %s",
					argv[2]);
				return (TCL_ERROR);
			}
			insflow(fp);
			fp->setfm(this);
			return (TCL_OK);
		}

	} else if (argc == 4) {
		if (strcmp(argv[1], "cat-drops-pthresh") == 0) {
			int category = atoi(argv[2]);
			int thresh = atoi(argv[3]);
			if (thresh < 0 || category > MAXCATEGORY) {
				tcl.resultf("flowmgr: invalid arr-thresh: %d",
					thresh);
				return (TCL_ERROR);
			}
			droptotals_[category].pkt_thresh_ = thresh;
			return (TCL_OK);
		}
		if (strcmp(argv[1], "cat-drops-bthresh") == 0) {
			int category = atoi(argv[2]);
			int thresh = atoi(argv[3]);
			if (thresh < 0 || category > MAXCATEGORY) {
				tcl.resultf("flowmgr: invalid arr-thresh: %d",
					thresh);
				return (TCL_ERROR);
			}
			droptotals_[category].byte_thresh_ = thresh;
			return (TCL_OK);
		}
                if (strcmp(argv[1], "cat-drops-pcallback") == 0) {
			int category = atoi(argv[2]);
                        const char* cb = argv[3];
			if (category < 0 || category > MAXCATEGORY) {
				tcl.resultf("flowmgr: invalid categ: %d",
					category);
				return (TCL_ERROR);
			}
                        delete droptotals_[category].pcallback_;
                        if (*cb == 0)
                                droptotals_[category].pcallback_ = 0;
                        else {
                                droptotals_[category].pcallback_ = new char[strlen(cb) + 1];
                                strcpy(droptotals_[category].pcallback_, cb);
                        }
                        return (TCL_OK);
                }
                if (strcmp(argv[1], "cat-drops-bcallback") == 0) {
			int category = atoi(argv[2]);
                        const char* cb = argv[3];
			if (category < 0 || category > MAXCATEGORY) {
				tcl.resultf("flowmgr: invalid categ: %d",
					category);
				return (TCL_ERROR);
			}
                        delete droptotals_[category].bcallback_;
                        if (*cb == 0)
                                droptotals_[category].bcallback_ = 0;
                        else {
                                droptotals_[category].bcallback_ = new char[strlen(cb) + 1];
                                strcpy(droptotals_[category].bcallback_, cb);
                        }
                        return (TCL_OK);
                }
	}
	tcl.resultf("ns flowmgr object: no such command %s", argv[1]);
	return (TCL_ERROR);
}


// ctors and dtors
FlowManager::FlowManager(u_int ftype) : ftype_(ftype),
    link_(0), nflows_(0), channel_(0),
    enable_arrivals_default_(1), enable_drops_default_(1),
    ndcat_(1), fwdrops_(0), fwarrivals_(0)
{
	memset(flows_, '\0', sizeof(flows_));
	memset(droptotals_, '\0', sizeof(droptotals_));
	memset(&alldrops_, '\0', sizeof(alldrops_));
	memset(&allarrivals_, '\0', sizeof(allarrivals_));
}

FlowManager::~FlowManager()
{
	freeflows();
}

//
// reset all counters associated with this flow manager
//	[i.e. resets all base stats plus all per-flow stats]
//
void
FlowManager::resetcounters()
{
	register Flow **p;
	register int i, j = ndcat_;
	for (i = 0; i < j; i++)
		droptotals_[i].zerocounters();
	alldrops_.zerocounters();
	allarrivals_.zerocounters();
	for (i = 0, p = flows_; i < nflows_; i++, p++)
		(*p)->resetcounters();
	fwdrops_ = 0;
	fwarrivals_ = 0;
	return;
}

// dump a particular category's flows
int
FlowManager::dumpflows(int categ)
{
	register Flow **p;
	register int i, rval = TCL_OK;

	for (i = 0, p = flows_; i < nflows_; i++, p++)
		if ((*p)->dump(channel_, categ) != TCL_OK)
			rval = TCL_ERROR;

	return (rval);
}

// dump all flows + categories
//	do it in this order:
//	flow1	cat0
//	flow1	cat1
//	...

int
FlowManager::dumpflows()
{
	register Flow **p;
	register int i, j, rval = TCL_OK;

	if (*flows_) {
		for (i = 0, p = flows_; i < nflows_; i++, p++) {
			for (j = 0; j < ndcat_; j++)
				if ((*p)->dump(channel_, j) != TCL_OK)
					rval = TCL_ERROR;
		}
	}

	return (rval);
}

Flow*
FlowManager::findflow(Packet* pkt)
{
	register Flow **p;
	register int i;

	for (i = 0, p = flows_; i < nflows_; i++, p++)
		if (match(pkt, *p))
			return(*p);

	return (0);
}

/* get rid of all the flows */
void
FlowManager::freeflows()
{
	register i;
	for (i = 0; i < nflows_; i++)
		delete flows_[i];
	nflows_ = 0;
}

void
FlowManager::remflow(Flow *f)
{
	register int i, slot;

	for (i = 0; i < nflows_; i++) {
		if (f == flows_[i]) {
			slot = i;
			break;
		}
	}
	if (i == nflows_) {
		fprintf(stderr, "FM:remflow: can't remove nonexist flow %p\n",
			f);
		return;
	}
	if (slot != nflows_-1)
		flows_[slot] = flows_[nflows_-1];
	nflows_--;
}

void
FlowManager::insflow(Flow *f)
{
	if (++nflows_ == MAXFLOWS) {
		fprintf(stderr, "exceeded max number of flows (%d)\n", MAXFLOWS);
		--nflows_;
		return;
	}
	flows_[nflows_-1] = f;
}


Flow *
FlowManager::newflow(Packet* pkt)
{
	if (++nflows_ == MAXFLOWS) {
		fprintf(stderr, "exceeded max number of flows (%d)\n", MAXFLOWS);
		--nflows_;
		return NULL;
	}
	Flow *f;

/* XXX support for RED "special" flows */
if (ftype_ == 0)
    f = flows_[nflows_-1] = new Flow(this,
	pkt->type_, pkt->class_, pkt->src_, pkt->dst_,
	enable_arrivals_default_, enable_drops_default_);
else if (ftype_ == FTYPE_RED)
    f = flows_[nflows_-1] = new RedFlow(this,
	pkt->type_, pkt->class_, pkt->src_, pkt->dst_,
	enable_arrivals_default_, enable_drops_default_);
else {
fprintf(stderr, "unknown ftype_ %d\n", ftype_);
}
	if (f == NULL)
		return NULL;
	return f;
}

/************************* flow class ****************************/

int Flow::uid_ = 0; // each flow of any type gets a unique id

//
// tcl interface to flow object:
//	$flow profile
//	$flow arrivals
//	$flow drops categ
int
Flow::command(int argc, const char*const* argv)
{
	// NB: the tcl code depends on the ordering of these sprintf's
        Tcl& tcl = Tcl::instance();
        if (argc == 2) {
                if (strcmp(argv[1], "profile") == 0) {  
			char *bp = tcl.buffer();
			sprintf(bp, "%u %u %u %u", 
				type_, class_, src_, dst_);
			tcl.result(bp);
                        return (TCL_OK);
		}

		if (strcmp(argv[1], "arrivals") == 0) {
			char *bp = tcl.buffer();
			sprintf(bp, "%u %u",
				arrivalinfo_.pkt_arrivals_,
				arrivalinfo_.byte_arrivals_);
			tcl.result(bp);
                        return (TCL_OK);
		}
		if (strcmp(argv[1], "drops") == 0) {
			char *bp = tcl.buffer();
			sprintf(bp, "%u %u",
				drops_.pkt_drops_,
				drops_.byte_drops_);
			tcl.result(bp);
                        return (TCL_OK);
		}
		if (strcmp(argv[1], "mark") == 0) {
			tcl.resultf("%d", mark_);
			return (TCL_OK);
		}
		if (strcmp(argv[1], "resetcounters") == 0) {
			resetcounters();
			return (TCL_OK);
		}
	} else if (argc == 3) {
		if (strcmp(argv[1], "drops") == 0) {
			int categ = atoi(argv[2]);
			if (categ < 0 || categ >= fm_->ndcat()) {
				tcl.resultf("flow drops: categ %d invalid\n",
					categ);
				return (TCL_ERROR);
			}
			char *bp = tcl.buffer();
			sprintf(bp, "%u %u",
				dropinfo_[categ].pkt_drops_,
				dropinfo_[categ].byte_drops_);
			tcl.result(bp);
                        return (TCL_OK);
		}
		if (strcmp(argv[1], "mark") == 0) {
			int mval = atoi(argv[2]);
			mark_ = mval;
			return (TCL_OK);
		}
	}
	return (TclObject::command(argc, argv));
}




// base class version of arrival():
//   log many stats associated with a packet arrival on this flow
void
Flow::arrival(Packet* pkt)
{

	int pkts, bytes, pthresh, bthresh;

	if (!fm_) {
		fprintf(stderr, "no flow manager link\n");
		return;
	}

	//
	// stats across all flows
	//

	fm_->allarrivals_.pkt_arrivals_++;			// across all flows
	fm_->allarrivals_.byte_arrivals_ += pkt->size_;		// across all flows
	if (arrivals_enabled_) {
		if (arrivalinfo_.pkt_arrivals_ == 0)
			fm_->fwarrivals_++;
		arrivalinfo_.pkt_arrivals_++;			// this flow only
		arrivalinfo_.byte_arrivals_ += pkt->size_;	// this flow only
	}

	pkts = fm_->allarrivals_.pkt_arrivals_;
	bytes = fm_->allarrivals_.byte_arrivals_;
	pthresh = fm_->allarrivals_.pkt_thresh_;
	bthresh = fm_->allarrivals_.byte_thresh_;

	if (pkts >= pthresh && pthresh > 0 && fm_->allarrivals_.pcallback_) 
		docallback(fm_->allarrivals_.pcallback_, NOCATEGORY);

	if (bytes >= bthresh && bthresh > 0 && fm_->allarrivals_.bcallback_)
		docallback(fm_->allarrivals_.bcallback_, NOCATEGORY);

	if (!arrivals_enabled_)
		return;

	//
	// stats local to this flow
	//


	pkts = arrivalinfo_.pkt_arrivals_;
	bytes = arrivalinfo_.byte_arrivals_;
	pthresh = arrivalinfo_.pkt_thresh_;
	bthresh = arrivalinfo_.byte_thresh_;

	if (pkts >= pthresh && pthresh > 0 && arrivalinfo_.pcallback_)
		docallback(arrivalinfo_.pcallback_, NOCATEGORY);

	if (bytes >= bthresh && bthresh > 0 && arrivalinfo_.bcallback_)
		docallback(arrivalinfo_.bcallback_, NOCATEGORY);

	return;
}



// base class version of drop():
//   drop a packet of a particular category, keep a number of stats
//   and if they exceed any specified thresholds, call a tcl handler
// Red.cc uses category 1 for unforced drops, and category 0 for
//   unforced drops.

void
Flow::drop(Packet* pkt, int category)
{
	int pkts, bytes, pthresh, bthresh;

	if (!fm_) {
		fprintf(stderr, "no flow manager link\n");
		return;
	}
	if (category > fm_->ndcat() || category > MAXCATEGORY || category < 0) {
		fprintf(stderr, "drop category %d invalid-- max [%d]\n",
			category, MAXCATEGORY);
		return;
	}


	fm_->droptotals_[category].pkt_drops_++;		// per cat, all flows
	fm_->droptotals_[category].byte_drops_ += pkt->size_;	// per cat, all flows
	fm_->alldrops_.pkt_drops_++;				// all, total aggregate pdrops
	fm_->alldrops_.byte_drops_ += pkt->size_;		// all, total aggregate bdrops
	if (drops_enabled_) {
		// per-flow
		if (drops_.pkt_drops_ == 0)
			fm_->fwdrops_++;
		dropinfo_[category].pkt_drops_++;
		dropinfo_[category].byte_drops_ += pkt->size_;
		drops_.pkt_drops_++;
		drops_.byte_drops_ += pkt->size_;
	}

	pkts = fm_->droptotals_[category].pkt_drops_;
	bytes = fm_->droptotals_[category].byte_drops_;
	pthresh = fm_->droptotals_[category].pkt_thresh_;
	bthresh = fm_->droptotals_[category].byte_thresh_;

	if (pkts >= pthresh && pthresh > 0 && fm_->droptotals_[category].pcallback_)
		docallback(fm_->droptotals_[category].pcallback_, category);

	if (bytes >= bthresh && bthresh > 0 && fm_->droptotals_[category].bcallback_)
		docallback(fm_->droptotals_[category].bcallback_, category);

	// all stats

	pkts = fm_->alldrops_.pkt_drops_;
	bytes = fm_->alldrops_.byte_drops_;
	pthresh = fm_->alldrops_.pkt_thresh_;
	bthresh = fm_->alldrops_.byte_thresh_;

	if (pkts >= pthresh && pthresh > 0 && fm_->alldrops_.pcallback_) {
		double now = Scheduler::instance().clock();
		if (fm_->alldrops_.min_time_ == 0.0 ||
		  now >= fm_->alldrops_.last_time_ + fm_->alldrops_.min_time_) {
			fm_->alldrops_.last_time_ = now;
			docallback(fm_->alldrops_.pcallback_, NOCATEGORY);
		}
	}

	if (bytes >= bthresh && bthresh > 0 && fm_->alldrops_.bcallback_)
		docallback(fm_->alldrops_.bcallback_, NOCATEGORY);

	if (!drops_enabled_)
		return;


	// per-flow stats

	pkts = dropinfo_[category].pkt_drops_;
	bytes = dropinfo_[category].byte_drops_;
	pthresh = dropinfo_[category].pkt_thresh_;
	bthresh = dropinfo_[category].byte_thresh_;

	if (pkts >= pthresh && pthresh > 0 && dropinfo_[category].pcallback_)
		docallback(dropinfo_[category].pcallback_, category);

	if (bytes >= bthresh && bthresh > 0 && dropinfo_[category].bcallback_)
		docallback(dropinfo_[category].bcallback_, category);

	pkts = drops_.pkt_drops_;
	bytes = drops_.byte_drops_;
	pthresh = drops_.pkt_thresh_;
	bthresh = drops_.byte_thresh_;

	if (pkts >= pthresh && pthresh > 0 && drops_.pcallback_)
		docallback(drops_.pcallback_, category);

	if (bytes >= bthresh && bthresh > 0 && drops_.bcallback_)
		docallback(drops_.bcallback_, category);

	return;
}

//
// flow resetcounters():
//	reset all counters local to a particular flow [no affect to fm counters]
//
void
Flow::resetcounters() {
	register i;
	register j = 0;
	if (fm_) {
		j = fm_->ndcat();
		for (i = 0; i < j; i++)
			dropinfo_[i].zerocounters();
	}
	arrivalinfo_.zerocounters();
	drops_.zerocounters();
	return;  
}

int
Flow::dump(Tcl_Channel tc, int categ)
{
	int n, k;
	if (tc == 0)
		return TCL_ERROR;

	n = formatc(categ);

	wrk_[n] = '\n';
	wrk_[n+1] = '\0';
	(void)Tcl_Write(tc, wrk_, n+1);
	wrk_[n] = '\0';

	return (TCL_OK);
}

int
Flow::formatc(int category)
{
/* #define DEBUGFLOW */
#ifdef DEBUGFLOW
	printf("Flow::format: [fid %d, cat %d]: parr: %d, barr: %d, pdrop: %d, bdrop: %d, total_parrivals:%d, total_barrivals: %d\n",
		fid, category,
		arrivalinfo_.pkt_arrivals_,
		arrivalinfo_.byte_arrivals_,
		dropinfo_[category].pkt_drops_,
		dropinfo_[category].byte_drops_,
		allarrivals_.pkt_arrivals_,
		allarrivals_.byte_arrivals_);
	fflush(stdout);

#endif
	// format [19 values]:
	//	time, flow-id, category, type, class, src, dst,
	//	arrivals-this-flow(pkts), arrivals-this-flow(bytes),
	//	drops-this-flow-and-cat(pkts), drops-this-flow-and-cat(bytes),
	//	all-arrivals(pkts), all-arrivals(bytes),
	//	all-drops-this-cat(pkts), all-drops-this-cat(bytes)
	//	all-drops(pkts), all-drops(bytes)
	//	drops-this-flow(pkts), drops-this-flow(bytes)

	sprintf(wrk_, "%8.3f %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d %d",
			Scheduler::instance().clock(), fid_, category,
			type_, class_, src_, dst_,
			arrivalinfo_.pkt_arrivals_,
			arrivalinfo_.byte_arrivals_,
			dropinfo_[category].pkt_drops_,
			dropinfo_[category].byte_drops_,
			fm_->allarrivals_.pkt_arrivals_,
			fm_->allarrivals_.byte_arrivals_,
			fm_->droptotals_[category].pkt_drops_,
			fm_->droptotals_[category].byte_drops_,
			fm_->alldrops_.pkt_drops_,
			fm_->alldrops_.byte_drops_,
			drops_.pkt_drops_,
			drops_.byte_drops_);
	return(strlen(wrk_));
}


//
// invoke a tcl callback procedure.   If no category is given, do:
//	callback-name link-ref fm-ref flow-ref
// if a category is given, do:
//	callback-name link-ref fm-ref flow-ref categ
//
void
Flow::docallback(char *callback, int categ)
{
	if (callback != 0) {
		Tcl& tcl = Tcl::instance();
		if (categ == NOCATEGORY) {
		    // eval("callback link_name fm_name flow_name")
		    tcl.evalf("%s %s %s %s",
			callback,
			fm_->link()->TclObject::name(),
			fm_->TclObject::name(),
			name());
		} else {
		    // eval("callback link_name fm_name flow_name categ")
		    tcl.evalf("%s %s %s %s %d",
			callback,
			fm_->link()->TclObject::name(),
			fm_->TclObject::name(),
			name(),
			categ);
		}
	} else {
		fprintf(stderr, "docallback(): null callback\n");
	}
	return;
}

/****************XXX red-specific ***********************************/

void
RedFlow::arrival(Packet* pkt)
{
	// stats local to this flow
	arrivalinfo_.byte_arrivals_ += pkt->size_;
	// stats across all flows
	fm_->allarrivals_.pkt_arrivals_++;
	fm_->allarrivals_.byte_arrivals_ += pkt->size_;
	return;
}

void
RedFlow::drop(Packet* pkt, int category) {
	if (!fm_) {
		fprintf(stderr, "no flow manager link\n");
		return;
	}
	if (category > fm_->ndcat() || category > MAXCATEGORY || category < 0) {
		fprintf(stderr, "drop category %d invalid-- max [%d]\n",
			category, MAXCATEGORY);
		return;
	}

dropinfo_[category].byte_drops_ += pkt->size_;
dropinfo_[category].pkt_drops_++;
drops_.byte_drops_ += pkt->size_;
drops_.pkt_drops_++;

#ifdef notdef
	// per-flow stats
        if (category == 0) {
		dropinfo_[category].byte_drops_ += pkt->size_;
        } else if (category == 1) {
		dropinfo_[category].pkt_drops_++;
        }
#endif
	// per-category stats
	fm_->droptotals_[category].byte_drops_ += pkt->size_;
	fm_->droptotals_[category].pkt_drops_++;
	// all cumulative stats
	fm_->alldrops_.pkt_drops_++;

	int pkts = fm_->alldrops_.pkt_drops_;
	int pthresh = fm_->alldrops_.pkt_thresh_;

	if (pkts >= pthresh && pthresh > 0 && fm_->alldrops_.pcallback_)
		// Notes on work in progress:
		// Here I need to call a procedure to munch on the
		// information.  But the procedure needs to be
		// in red.cc, because I need information about 
		// the average queue size, as a measure of the level
		// of congestion.  I need to pass the information
		// from dump() to red.cc.  
		// First, the fraction of drops that were forced.
		// Second, for each flowID with drops, the single
		// measure of bandwidth for that flowID:
		//  (dropinfo_[0].byte_drops_ / droptotals_[0].byte_drops_) 
		//    * droptotals_[0].pkt_drops_ / alldrops_.pkt_drops_
		// + (dropinfo_[1].pkt_drops_ / droptotals_[1].pkt_drops_) 
		//  * droptotals_[1].pkt_drops_ / alldrops_.pkt_drops_
		docallback(fm_->alldrops_.pcallback_, NOCATEGORY);

	return;
}
