/*
 * Copyright (c) 1996 The Regents of the University of California.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 * 	This product includes software developed by the Network Research
 * 	Group at Lawrence Berkeley National Laboratory.
 * 4. Neither the name of the University nor of the Laboratory may be used
 *    to endorse or promote products derived from this software without
 *    specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

//
// FlowManager classes and a Flow class
//	idea is that a routing entity may create a flow manager
//	and pass packets to it.  The manager will keep track of
//	various statistics regarding the flow
//
// This is presently experimental
//

#ifndef ns_flow_h
#define ns_flow_h

#include <stdio.h>
#include "object.h"
#include "packet.h"

class Tcl;
class Link;

//
// a flow manager-- one of these per node interested in monitoring flows
//
// for now, the flows are maintained as a simple array-- if the
// number of flows per node is actually fairly large,
// this should become a hash table or other structure capable
// of faster retrieval (kf)
//
#define	MAXFLOWS	2048	/* max # flows at one node */
#define	MAXCATEGORY	3	/* max # of "categories" of drops */
#define	NOCATEGORY	(-1)

// information pertaining to dropped packets
struct drpinfo {
	u_int	pkt_drops_;	/* # packets dropped */
	u_int	byte_drops_;	/* # bytes in dropped packets */
	u_int	pkt_thresh_;	/* threshold; invoke pcallback if exceeded */
	u_int	byte_thresh_;	/* threshold; invoke bcallback if exceeded */
	char*	pcallback_;	/* tcl callback */
	char*	bcallback_;	/* tcl callback */
	double  min_time_;	/* minimum time between callbacks */
	double  last_time_;	/* time of last callback */
	inline void zerocounters() {
		pkt_drops_ = byte_drops_ = 0;
	}

};

// information pertaining to arriving packets (see above)
struct arrinfo {
	u_int	pkt_arrivals_;
	u_int	byte_arrivals_;
	u_int	pkt_thresh_;
	u_int	byte_thresh_;
	char*	pcallback_;
	char*	bcallback_;
	inline void zerocounters() {
		pkt_arrivals_ = byte_arrivals_ = 0;
	}
};

//
// flow structure-- fields storing which corresponding packet header field
//	values must match to be on the same "flow".  the flow stucture is
//	used across all possible sorts of flows, where only the match()
//	function in the flow managers (above) is different
//	Also, this structure holds per-flow accumulated stats

class FlowManager;
class Flow : public TclObject {
	static int	uid_;
    private:
	int formatc(int);
    protected:
	int		fid_;
	u_short		src_;
	u_short		dst_;
	u_int		type_;
	u_int		class_;

	u_int8_t	arrivals_enabled_;
	u_int8_t	drops_enabled_;
	int		mark_;	// defined by application
	FlowManager*	fm_;	// back-pointer to associated fm


	// per-flow stats
	struct drpinfo		dropinfo_[MAXCATEGORY];		// per-cat
	struct arrinfo		arrivalinfo_;			// arrivals
	struct drpinfo		drops_;				// drops

	char			wrk_[256];

	void docallback(char*, int);
	int command(int argc, const char*const* argv);
    public:
	Flow(FlowManager *fm, u_int t, u_int cl, u_short s, u_short d,
	  int ea, int ed) :
	    fid_(++uid_), src_(s), dst_(d), type_(t), class_(cl),
	    arrivals_enabled_(ea), drops_enabled_(ed), fm_(fm), mark_(0)
	{
		memset(dropinfo_, '\0', sizeof(dropinfo_));
		memset(&arrivalinfo_, '\0', sizeof(arrivalinfo_));
		memset(&drops_, '\0', sizeof(drops_));
		wrk_[0] = 0;
		/* type_, class_, src_, dst_ will be filled in shortly */
	}
	void resetcounters();

	u_short src() { return (src_); }
	u_short dst() { return (dst_); }
	u_int type() { return (type_); }
	u_int fclass() { return (class_); }
		// can't be class(), a reserved word in C++

	virtual void drop(Packet* pkt, int category = 0);
	virtual void arrival(Packet* pkt);

	inline u_int pdrops(int cat) { return(dropinfo_[cat].pkt_drops_); }
	inline u_int bdrops(int cat) { return(dropinfo_[cat].byte_drops_); }
	inline u_int pdrops() { return(drops_.pkt_drops_); }
	inline u_int bdrops() { return(drops_.byte_drops_); }
	inline u_int parrivals() { return(arrivalinfo_.pkt_arrivals_); }
	inline u_int barrivals() { return(arrivalinfo_.byte_arrivals_); }
	inline void setfm(FlowManager *fmp) { fm_ = fmp; }

	int dump(Tcl_Channel, int);
};

class RedFlow : public Flow {
    public:
	inline RedFlow(FlowManager *fm, u_int t, u_int cl, u_short s, u_short d, int ea, int ed) :
	Flow(fm, t, cl, s, d, ea, ed) { }
	void drop(Packet* pkt, int category = 0);
	void arrival(Packet* pkt);
};


#define	FTYPE_RED	1	// XXX, used to identify special RED machinery

class FlowManager : public TclObject {
    friend void Flow::drop(Packet*, int);
    friend void Flow::arrival(Packet*);
    friend void RedFlow::drop(Packet*, int);
    friend void RedFlow::arrival(Packet*);
    friend int Flow::formatc(int);

    protected:
	u_int ftype_;		// XXX this is temporary for RED support
	/* flow management */
	Link* link_;		// back-ptr to referring link
	u_int nflows_;		// # of current flows
	Tcl_Channel channel_;	// file for trace data
	Flow* flows_[MAXFLOWS];	// flow information
	int enable_arrivals_default_;	// new flows keep arrival info?
	int enable_drops_default_;	// new flows keep drop info?

	/* data common across all flows associated w/this manager */
	u_int 		ndcat_;				// # of drop categs
	u_int		fwdrops_;			// # flows with >0 drops
	u_int		fwarrivals_;			// # flows with >0 arrivals
	struct drpinfo	droptotals_[MAXCATEGORY];	// total-drops-per-cat
	struct drpinfo	alldrops_;			// drops per fm
	struct arrinfo	allarrivals_;			// arrivals per fm
	double computeFrac(int, int);

    public:
	FlowManager(u_int ftype = 0);
	~FlowManager();
	int command(int argc, const char*const* argv);
	Flow* findflow(Packet*);	// look up a flow
	virtual int match(Packet*, Flow*) = 0;
	Flow* newflow(Packet*);		// create new flow
	inline void link(Link *lk) { link_ = lk; };
	inline Link* link() { return(link_); };
	void resetcounters();		// reset all drpinfo/arrinfo counters
	int dumpflows(int);	// dump flows in a category
	int dumpflows();	// dump all flows
	inline u_int ndcat() { return (ndcat_); }
	inline u_int fwdrops() { return (fwdrops_); }
	inline u_int fwarrivals() { return (fwarrivals_); }
	inline u_int total_pdrops(int ct) { return(droptotals_[ct].pkt_drops_); }
	inline u_int total_bdrops(int ct) { return(droptotals_[ct].byte_drops_); }
	inline u_int total_parrivals() { return(allarrivals_.pkt_arrivals_); }
	inline u_int total_barrivals() { return(allarrivals_.byte_arrivals_); }
	void insflow(Flow *);
	void remflow(Flow *);
	void freeflows();
};



class ClassFlowManager : public FlowManager {
    protected:
	inline int match(Packet* pkt, Flow *fp) {
		return (pkt->class_ == fp->fclass());
	};
};

class SrcFlowManager : public FlowManager {
    protected:
	inline int match(Packet* pkt, Flow *fp) {
		return (pkt->src_ == fp->src());
	};
};

class DstFlowManager : public FlowManager {
    protected:
	inline int match(Packet* pkt, Flow *fp) {
		return (pkt->dst_ == fp->dst());
	};
};

class SrcDstFlowManager : public FlowManager {
    protected:
	inline int match(Packet* pkt, Flow *fp) {
		return (pkt->src_ == fp->src() &&
			pkt->dst_ == fp->dst());
	};
};

class SrcDstClassFlowManager : public FlowManager {
    public:
	inline SrcDstClassFlowManager(u_int ftype = 0) : FlowManager(ftype) { }
    protected:
	inline int match(Packet* pkt, Flow *fp) {
		return (pkt->src_ == fp->src() &&
			pkt->dst_ == fp->dst() &&
			pkt->class_ == fp->fclass());
	};
};

#endif
