/*                                                                           */
/*      Copyright (c) 1997 by the University of Southern California          */
/*      All rights reserved.                                                 */
/*                                                                           */
/*      Permission to use, copy, modify, and distribute this software and    */
/*      its documentation in source and binary forms for any purpose and     */
/*      without fee is hereby granted, provided that both the above          */
/*      copyright notice and this permission notice appear in all copies,    */
/*      and that any documentation, advertising materials, and other         */
/*      materials related to such distribution and use acknowledge that      */
/*      the software was developed in part by the University of Southern     */
/*      California, Information Sciences Institute.  The name of the         */
/*      University may not be used to endorse or promote products derived    */
/*      from this software without specific prior written permission.        */
/*                                                                           */
/*                                                                           */
/*      THE UNIVERSITY OF SOUTHERN CALIFORNIA makes no representations about */
/*      the suitability of this software for any purpose.  THIS SOFTWARE IS  */
/*      PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR IMPLIED WARRANTIES,      */
/*      INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF             */
/*      MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.                */
/*                                                                           */

#ifndef lint
static char rcsid[] =
    "@(#) $Header: rtglib.cc,v 1.2 97/01/23 23:26:10 kfall Exp $ (USC/ISI)";
#endif

#include <stdlib.h>
#include "Tcl.h"
#include "object.h"
#include "scheduler.h"
#include "packet.h"
#include "agent.h"
#include "node.h"
#include "link.h"
#include "rtglib.h"

static class rtObjectMatcher : public Matcher {
public:
	rtObjectMatcher() : Matcher("agent") {}
	TclObject*	match(const char *id) {
		if (strcasecmp(id, "rtobject") == 0) {
			return new rtObject;
		}
		return NULL;
	}
} matcher_rtobject;

//
// This file manipulates the rtObject, rtProto, and rtPeer classes.
// rtPeer is currently only nominally an object class.  All of its
// elements are public and encapsulated inside the rtProto class.
//
// The relationship between the various route structures is:
//
// 	rtObject
// 	{
// 		rts_ { /* Loc-RIB */   // value when dest. unreachable VVV
// 			cost_;	// of this route,	      [DVInfinity]
// 			pref_;	// of this route, RW,	      [-1]
// 			if_;	// that this route uses,      [NULL]
// 		}[#destinations]
// 	}
//
// 	rtProto
// 	{
// 		pref_;		// global preference for protocol, RO
// 		rts_ {		// best routes in this protocol
// 			cost_;
// 			pref_;	// current pref. of this route, RW
// 			peer_;
// 		}[#destinations];
// 	}
//
// 	rtPeer 
// 	{
// 		addr_, port_;	// identifies the peer
// 		pref_;		// pref. for routes from this peer, RO
// 		rtsin_ { /* Adj-RIB-In from this peer */
// 			cost_;	// learned from peer
// 			pref_;	// of this route from this peer, RO
// 		}[#destinations];
// 		rtsout_[#destinations];
// 	}
//
// Variables/preferences marked RW are those that are changed in these
// objects.  Those marked RO are set by the appropriate init
// functions, and modified through interpreted operations.
//


rtObject::rtObject() : Agent(-1),
	rts_(0), rtproto_(0), linkstate_(0), proto_rts_changed_(0),
	time_up_(0), time_dn_(0)
{
}
	

rtObject::~rtObject() 
{
	Link_State* lsp;
	rtProto* pr;

	for (lsp = linkstate_; lsp ; /* NOTHING */) {
		lsp = lsp->next_;
		delete lsp;
	}
	for (pr = rtproto_; pr; /* NOTHING */) {
		pr = pr->next_;
		delete pr;
	}
	delete rts_;
}


int rtObject::command(int argc, const char*const* argv)
/*
 * $rtobject attach $node
 */
{
	if (argc == 2) {
		if (strcmp(argv[1], "init") == 0) {
			init();
			return TCL_OK;
			
		}
		if (strcmp(argv[1], "notify-ifchanged") == 0) {
			intf_changed();
			return TCL_OK;
		}
	}
	if (argc == 3) {
		if (strcmp(argv[1], "dump-routes") == 0) {
			Trace* t = (Trace*) TclObject::lookup(argv[2]);
			if (t) {
				dump_routes(t);
				return TCL_OK;
			} else {
				Tcl& tcli = Tcl::instance();
				tcli.resultf("dump-routes %s: cannot find trace object %s",
					     argv[0], argv[2]);
				return TCL_ERROR;
			}
		}
	}
	return Agent::command(argc, argv);
}


void rtObject::add_proto(rtProto* proto)
{
	proto->next_ = rtproto_;
	rtproto_ = proto;
}


void rtObject::init()
{
	register Link* lp;
	register int i;

	for (lp = node_->links(); lp; lp = lp->next_) {
		Link_State* lsp = new Link_State;
		lsp->link_ = lp;
		lsp->state_ = 1;
		lsp->next_ = linkstate_;
		linkstate_ = lsp;
	}
	rts_ = new RIB[node_->cnt()];
	proto_rts_changed_ = 0;
	for (i = 0; i < node_->cnt(); i++) {
		rts_[i].if_ = (Link*) NULL;
		rts_[i].peer_ = (rtPeer*) NULL;
		rts_[i].proto_ = (rtProto*) NULL;
		rts_[i].cost_ = DVInfinity;
		rts_[i].pref_ = MINPREF;
	}
	rts_[node_->addr()].cost_ = MINCOST;
	rts_[node_->addr()].pref_ = MAXPREF;
	node_->setroute(node_->addr(), (Link*) NULL);
	for (lp = node_->links(); lp; lp = lp->next_) {
		i = lp->dst();
		rts_[i].if_ = lp;
		// need a direct rtproto...
		rts_[i].cost_ = 1;
		rts_[i].pref_ = 100;
		node_->setroute(i, lp);
	}
	time_up_ = Scheduler::instance().clock();
}


void rtObject::intf_changed()
/*
 * Called when LS_QUERY_TIMER fires.
 *	1. Check link states on node.
 *	2. if link states have changed,
 *		notify rtproto to recompute RIBs
 *	3. if rtproto flags proto RIBIn changes,
 *		update the FIB (Forwarding Information base)
 *		notify rtproto to send updates to all peers.
 */
{
	register i;
	rtProto *pr;

	if (check_link_states()) {
		for (i = 0; i < node_->cnt(); i++)
			if (rts_[i].if_ && rts_[i].if_->down()) {
				node_->setroute(i, (Link*) NULL);
				rts_[i].if_ = (Link*) NULL;
				rts_[i].peer_ = (rtPeer*) NULL;
				rts_[i].proto_ = (rtProto*) NULL;
				rts_[i].cost_ = DVInfinity;
				rts_[i].pref_ = MINPREF;
				time_dn_ = Scheduler::instance().clock();
			}
		for (pr = rtproto_; pr; pr = pr->next_)
			pr->compute_RIBs();
		compute_LocRIB();
	}
	update_protos();
}


inline int rtObject::check_link_states()
{
	int retval;
	Link_State *lsp;
	for (lsp = linkstate_; lsp; lsp = lsp->next_)
		if (lsp->state_ ^ lsp->link_->state()) {
			lsp->state_ = lsp->link_->state();
			retval = 1;
		}
	return retval;
}

inline int rtObject::BetterRoute(const RIB *r1, const RIB *r2)
// true if r1 is a Better Route than r2 
{
	return (r1->if_ &&			  // have a route and
		((r1->pref_ > r2->pref_) ||	  // of better preference or
						  // equal pref. and lower cost
		 ((r1->pref_ == r2->pref_) && (r1->cost_ < r2->cost_))));
}


void rtObject::compute_LocRIB()
/*
 * If any protocol routes have changed,
 *	for every destination,
 *		compute the best route learned from any routing protocol.
 *	if new routes are found
 *		install_FIB()
 *		notify send updates
 */
{
	rtProto* pr;
	register i;

	if (proto_rts_changed_) {
		// find best routes from each protocol instance
		int rtchanged = 0;
		for (pr = rtproto_; pr; pr = pr->next_)
			for (i = 0; i < node_->cnt(); i++)
				if (BetterRoute(pr->rts(i), &rts_[i])) {
					rts_[i].proto_ = pr;
					rts_[i].if_   = pr->rts(i)->if_;
					rts_[i].cost_ = pr->rts(i)->cost_;
					rts_[i].pref_ = pr->rts(i)->pref_;
					rtchanged = 1;
				}
		// install FIB; notify send updates
		if (rtchanged) {
			for (i = 0; i < node_->cnt(); i++)
				node_->setroute(i, rts_[i].if_);
			for (pr = rtproto_; pr; pr = pr->next_)
				pr->notify_send_updates();
		}
		time_up_ = Scheduler::instance().clock();
		proto_rts_changed_ = 0;
	}
}


void rtObject::update_protos() 
{
	rtProto* pr;
	for (pr = rtproto_; pr; pr = pr->next_)
		pr->send_update();
}


void rtObject::dump_routes(Trace* t) 
{
	register i;
	if (rts_ == NULL) {
		Tcl::instance().result("ns: testbed is not initialised");
		return;
	}
	
	sprintf(t->buffer(),
		"node %d:\n\tdest.\tnexthop\tcost\tpref.\tproto\ttime",
		node_->addr());
	t->dump();
	for (i = 0; i < node_->cnt(); i++) {
		if (rts_[i].if_) {
			sprintf(t->buffer(), "\t%d\t%d\t%d\t%d",
				i, rts_[i].if_->dst(),
				rts_[i].cost_, rts_[i].pref_);
			if (rts_[i].proto_) {
				sprintf(t->buffer(), "%s\t%s\t%g",
					t->buffer(),
					rts_[i].proto_->name(),
					rts_[i].proto_->time());
			} else {
				sprintf(t->buffer(), "%s\t%s",
					t->buffer(), "direct");
			}
		} else if (i == node_->addr()) {
			sprintf(t->buffer(), "\t%d\t%d\t%d\t%d\t%s",
				i, i, 0, MAXPREF, "local");
		} else {
			sprintf(t->buffer(), "\t%d\tDOWN", i);
		}
		t->dump();
	}
	sprintf(t->buffer(), "time last route install %g\tuninstall %g\n",
		time_up_, time_dn_);
	t->dump();
}


//////////////////////////////// rtProto ///////////////////////////////////

static class rtProtoMatcher : public Matcher {
public:
	rtProtoMatcher() : Matcher("agent") {}
	TclObject*	match(const char *id) {
		if (strcasecmp(id, "rtproto") == 0) {
			return new rtProto;
		}
		return NULL;
	}
} matcher_rtproto;


rtProto::rtProto() : Agent(PT_ROUTE_DV), name_(0), rtobj_(0),
	peers_(0), rts_(0), peer_rts_changed_(0), time_(0)
{
	link_int("ns_rtconf", "pref-dv", &pref_, 0);
}


rtProto::~rtProto() 
{
	rtPeer* peer;
	delete name_;
	for (peer = peers_; peer; /* NOTHING */) {
		peer = peer->next_;
		delete peer->rtsin_;
		delete peer->rtsout_;
		delete peer;
	}
	delete rts_;
}


int rtProto::command(int argc, const char*const* argv)
/*
 * $rtproto proto <prototype>		# currently only "DV"
 * $rtproto rtagent $rtobject
 * $rtproto pref <value> default
 * $rtproto pref <value> peer <address>
 * $rtproto pref <value> peer <address> destination <address>
 */
{
	if (argc == 3) {
		if (strcasecmp(argv[1], "proto") == 0) {
			name_ = new char[strlen(argv[2]) + 1];
			strcpy(name_, argv[2]);
			return TCL_OK;
		}
		if (strcasecmp(argv[1], "rtagent") == 0) {
			rtobj_ = (rtObject*) TclObject::lookup(argv[2]);
			rtobj_->add_proto(this);
			init();
			return TCL_OK;
		}
	}
	if (argc == 4) {
		if (strcasecmp(argv[1], "peer-init") == 0) {
			Link* lp = (Link*) TclObject::lookup(argv[2]);
			rtProto* pp = (rtProto*) TclObject::lookup(argv[3]);
			peerinit(lp, pp);
			return TCL_OK;
		}
	}
	if (strcasecmp(argv[1], "pref") == 0) {
		// XXX This code is gunky and very poorly thought out.
		int p = atoi(argv[2]);
		if (p > MAXPREF)
			p = MAXPREF;
		if (strcasecmp(argv[3], "default") == 0) {
			pref_ = p;
		} else if (strcasecmp(argv[3], "peer") == 0) {
			rtPeer* peer = get_peer(atoi(argv[4]));
			if (! peer) {
				Tcl::instance().resultf("peer %s %s %s",
							argv[3], 
							"not found in",
							argv[0]);
				return TCL_ERROR;
			}
// 			if (argc == 7 &&
// 			    strcasecmp(argv[5], "destination") == 0) {
// 				int dest = atoi(argv[6]);
			peer->pref_ = atoi(argv[4]);
		}
		return TCL_OK;
	}
	return Agent::command(argc, argv);
}


void rtProto::init()
{
	register i;
	peer_rts_changed_ = 0;
	rts_ = new RIB[node_->cnt()];
	for (i = 0; i < node_->cnt(); i++) {
		rts_[i].if_ = (Link*) NULL;
		rts_[i].peer_ = (rtPeer*) NULL;
		rts_[i].proto_ = this;
		rts_[i].cost_ = DVInfinity;
		rts_[i].pref_ = INVPREF;
	}
	sched(0, DV_REFRESH_TIMER);
}

void rtProto::peerinit(Link* lp, rtProto* rp)
{
	rtPeer* peer = new rtPeer;
	peer->rtproto_ = this;
	peer->addr_ = rp->addr();
	peer->port_ = rp->port();
	peer->if_ = lp;
	peer->state_ = 0;
	peer->send_update_ = 0;
	peer->pref_ = INVPREF;
	peer->rtsin_ = new RIB[node_->cnt()];
	peer->rtsout_ = new char[node_->cnt()];
	int i;
	for (i = 0; i < node_->cnt(); i++) {
		peer->rtsin_[i].if_ = (Link*) NULL;
		peer->rtsin_[i].peer_ = peer;
		peer->rtsin_[i].proto_ = this;
		peer->rtsin_[i].cost_ = DVInfinity;
		peer->rtsin_[i].pref_ = INVPREF;
		peer->rtsout_[i] = DVInfinity;
	}
	peer->next_ = peers_;
	peers_ = peer;
}

				
void rtProto::compute_RIBs()
/*
 * update link states, since the time they were last examined.
 * 	If the link has gone down recently,
 *		mark peer AdjRIBIns down
 *		If any route is best for that protocol,
 *			mark that route down,
 *			notify recompute proto RIBs
 *	else if the link has come up recently
 *		notify send update on that link
 * if peer routes have changed
 *	// they can change either due to link state change, or
 *	// routes received from a peer
 *	compute best routes for that proto
 *	If protocol RIBs have changed
 *		flag recompute LocRIB
 */
{
	rtPeer* peer;
	for (peer = peers_; peer; peer = peer->next_)
		if (peer->state_ && peer->if_ && peer->if_->down()) {
			// Line is *now* down
			peer->state_ = 0;
			int i;
			for (i = 0; i < node_->cnt(); i++) {
				peer->rtsin_[i].cost_ = DVInfinity;
				peer->rtsout_[i] = DVInfinity;
				if (rts_[i].if_ != peer->if_)
					continue;
				// else adjust the proto AdjRIBIn entry
				rts_[i].if_ = (Link*) NULL;
				rts_[i].cost_ = DVInfinity;
				rts_[i].pref_ = MINPREF;
				peer_rts_changed_ = 1;
			}
		} else if (!peer->state_ && peer->if_ && peer->if_->up()) {
			// line is *now* up
			peer->state_ = 1;
			peer->rtsin_[peer->if_->dst()].cost_ = 1;
			peer_rts_changed_ = 1;
			peer->send_update_ = 1;
		}
	if (peer_rts_changed_) {
		rtPeer* peer;
		for (peer = peers_; peer; peer = peer->next_) {
			int i;
			for (i = 0; i < node_->cnt(); i++) {
				int p = get_pref(&peer->rtsin_[i]);
				if (BetterRoute(&peer->rtsin_[i], p, &rts_[i])) {
					rts_[i].if_ = peer->if_;
					rts_[i].peer_ = peer->rtsin_[i].peer_;
					rts_[i].cost_ = peer->rtsin_[i].cost_;
					rts_[i].pref_ = p;
					rtobj_->mark_routes_changed();
					time_ = Scheduler::instance().clock();
				}
			}
		}
		peer_rts_changed_ = 0;
	}
}


inline int rtProto::get_pref(const RIB* r1)
{
	return ((r1->pref_ >= 0) ? r1->pref_ :
		 (r1->peer_->pref_ >= 0) ? r1->peer_->pref_ :
		r1->peer_->rtproto_->pref_);
}
	
	
inline int rtProto::BetterRoute(const RIB* r1, const int pref1, const RIB* r2)
/*
 * r1 is a better route than r2 if
 *	1.  r1 is a valid cost, and r1 has a higher preference than r2
 *	2.  r1 is a lower cost route of equal preference, or
 *	3.  r2 does not have a valid cost, and r1 does.
 */
{
	return (((r1->cost_ < DVInfinity) && (pref1 > r2->pref_)) ||
		((r1->cost_ < r2->cost_) && (pref1 == r2->pref_)) ||
		((r2->cost_ == DVInfinity) && (r1->cost_ < DVInfinity)));
}


void rtProto::timeout(int tno)
{
	compute_RIBs();
	notify_send_updates();
	rtobj_->compute_LocRIB();
	rtobj_->update_protos();
	sched(DVRefreshTime, tno);
}


void rtProto::recv(Packet *pkt)
/*
 * find the peer that the message is from
 * update the peer AdjRIBIns
 * If routes have changed
 *	compute RIBs
 *	send_update
 */
{
	rtPeer*	peer = get_peer(pkt->src_);
	char*	msg = (char*)pkt->bd_.msg_;
	
	if (! peer) {
		fprintf(stderr, "ns at %d: unknown peer for proto DV: %d\n", pkt->dst_, pkt->src_);
	} else {
		int rtchanged = 0;
		int i;
		for (i = 0; i < node_->cnt(); i++) {
			char ncost = *msg++;
			if (ncost < DVInfinity)
				ncost++;
			if (peer->rtsin_[i].cost_ != ncost) {
				peer->rtsin_[i].cost_ = ncost;
				peer_rts_changed_ = 1;
			}
		}
			
	}
	Packet::free(pkt);
	compute_RIBs();
	rtobj_->compute_LocRIB();
	rtobj_->update_protos();
}


inline void rtProto::notify_send_updates() 
{
	rtPeer* peer;
	for (peer = peers_; peer; peer = peer->next_)
		peer->send_update_ = 1;
}


inline void rtProto::send_update()
{
	rtPeer* p;
	for (p = peers_; p; p = p->next_)
		if (p->send_update_) {
			send_update(p);
			p->send_update_ = 0;
		}
}

void rtProto::send_update(rtPeer* p)
{
	if (p->if_->down())
		return;
	
	int n = node_->cnt();
	Packet* pkt = allocpkt();
	char*	msg = (char*)pkt->bd_.msg_;
	
	register i;
	for (i = 0; i < n; i++) {
		p->rtsout_[i] = *msg++ = ((rtobj_->rts(i)->if_ == p->if_) ?
					  DVInfinity : rtobj_->rts(i)->cost_);
	}
	pkt->size_  = sizeof(char) * n;
	pkt->dst_   = p->addr_;
	pkt->dport_ = p->port_;
	node_->transmit(pkt);
}
