#include "agent.h"
#include "srm-ctl.h"
#include "srm.h"

class SRMReq : public SRMObj {
public:
	SRMReq(int state, int seqnum) : SRMObj(0.0,state,0.0,seqnum,0) {};

	virtual void recv_repair(SRMCtl *ctl);
	virtual void recv_nack(SRMCtl *ctl);
	virtual void expire(SRMCtl *ctl);
};

void SRMReq::recv_repair(SRMCtl *ctl)
{
	// fprintf(stderr,"recv_repair: %p %p %d %d\n",ctl,this,m_state,m_seqnum);
	if (m_state == SRM_PENDING_REQUEST) {
		// we have been requesting it, now transition
		ctl->cancel(m_seqnum);

		m_state = SRM_REPAIRED;
		double delay;

		ctl->agent()->repaired_hook(1);
		
		// set the ignore delay
		delay = ctl->agent()->ignore_tmo();

		ctl->sched(this,delay);		
	} else if (m_state == SRM_REPAIRED) {
		// do nothing
	} else {
		fprintf(stderr,"Bad state for SRM\n");
	}
}

void SRMReq::recv_nack(SRMCtl *ctl)
{
	// fprintf(stderr,"recv_nack: %p %p %d %d\n",ctl,this,m_state,m_seqnum);
	if (m_state == SRM_PENDING_REQUEST) {
		// we have been requesting it, now transition
		ctl->cancel(m_seqnum);

		double delay;

		// Set the delay
		delay = ctl->agent()->backoff_tmo(m_delay);

		ctl->sched(this,delay);		
	} else if (m_state == SRM_REPAIRED) {
		// do nothing
	} else {
		fprintf(stderr,"Bad state for SRM\n");
	}
}

void SRMReq::expire(SRMCtl *ctl)
{
	// fprintf(stderr,"expire: %p %p %d %d\n",ctl,this,m_state,m_seqnum);
	if (m_state == SRM_PENDING_REQUEST) {
		double delay;

		// this is where we send the NACK!!!
		ctl->agent()->send_nack(m_seqnum);
		
		// set the delay
		delay = ctl->agent()->backoff_tmo(m_delay);
		ctl->sched(this,delay);
	} else if (m_state == SRM_REPAIRED) {
		// do nothing, just delete ourselves
		delete this;
	} else {
		fprintf(stderr,"Bad state for SRM\n");
	}
}

class SRMRep : public SRMObj {
public:
	SRMRep(int state, int seqnum) : SRMObj(0.0,state,0.0,seqnum,0) {};

	virtual void recv_repair(SRMCtl *ctl);
	virtual void recv_nack(SRMCtl *ctl);
	virtual void expire(SRMCtl *ctl);
};

void SRMRep::recv_repair(SRMCtl *ctl)
{
	// fprintf(stderr,"recv_repair: %p %p %d %d\n",ctl,this,m_state,m_seqnum);
	if (m_state == SRM_PENDING_REPAIR) {
		// It has been repaired by someone else
		ctl->cancel(m_seqnum);

		ctl->agent()->repaired_hook(0);

		m_state = SRM_REPAIRED;
		double delay;

		// set the ignore delay
		delay = ctl->agent()->ignore_tmo();

		ctl->sched(this,delay);
	} else if (m_state == SRM_REPAIRED) {
		// do nothing
	} else {
		fprintf(stderr,"Bad state for SRM\n");
	}

}

void SRMRep::recv_nack(SRMCtl *ctl)
{
	// fprintf(stderr,"recv_nack: %p %p %d %d\n",ctl,this,m_state,m_seqnum);
	if (m_state == SRM_PENDING_REPAIR) {
		// do nothing
	} else if (m_state == SRM_REPAIRED) {
		// ignore
	} else {
		fprintf(stderr,"Bad state for SRM\n");
	}
}

void SRMRep::expire(SRMCtl *ctl)
{
	// fprintf(stderr,"expire: %p %p %d %d\n",ctl,this,m_state,m_seqnum);
	if (m_state == SRM_PENDING_REPAIR) {
		m_state = SRM_REPAIRED;
		double delay;

		ctl->agent()->repaired_hook(0);

		// this is where we send the repair
		ctl->agent()->send_repair(m_seqnum);
		
		// set the ignore delay
		delay = ctl->agent()->ignore_tmo();

		ctl->sched(this,delay);		
	} else if (m_state == SRM_REPAIRED) {
		// just delete ourselves
		delete this;
	} else {
		fprintf(stderr,"Bad state for SRM\n");
	}
}
	
SRMCtl::SRMCtl(void)
	: m_list(0), m_agent(0), m_expected_seqnum(-1)
{
}

SRMCtl::~SRMCtl(void)
{
	SRMObj *out;
	
	while (m_list != 0) {
		out = m_list;
		m_list = m_list->m_next;
		delete out;
	}
}

void SRMCtl::reset(void)
{
	SRMObj *out;
	
	while (m_list != 0) {
		out = m_list;
		m_list = m_list->m_next;
		delete out;
	}
	m_expected_seqnum = -1;
}

SRMObj *SRMCtl::cancel(int seqnum)
{
	SRMObj *prev, *curr;

	prev = curr = 0;
	for (curr = m_list; curr != 0 && curr->m_seqnum != seqnum;
		 curr = curr->m_next)
		prev = curr;

	if (curr != 0) {
		if (prev == 0)
			m_list = curr->m_next;
		else
			prev->m_next = curr->m_next;

		return curr;
	}
	return 0;
}

void SRMCtl::sched(SRMObj *ob, double delay)
{
	double curr_tm = Scheduler::instance().clock();
	double expire;

//	if (find(ob->m_seqnum) != 0) {
//		printf("Already in %d,%d - %d,%d\n",ob->m_state,ob->m_seqnum,
//			   find(ob->m_seqnum)->m_state,find(ob->m_seqnum)->m_seqnum);
//	}
	
	expire = curr_tm + delay;
	
	ob->m_expire = expire;
	ob->m_delay = delay;

//	fprintf(stderr,"sched - %p %p %d %d\n",this,ob,ob->m_state,ob->m_seqnum);
	
	if (m_list == 0 || (expire < m_list->m_expire)) {
		ob->m_next = m_list;
		m_list = ob;
	} else {
		// Have to hunt for it!
		SRMObj *prev, *after;

		prev = m_list;
		after = m_list->m_next;

		while (after != 0 && (expire > after->m_expire)) {
			prev = after;
			after = after->m_next;
		}
		prev->m_next = ob;
		ob->m_next = after;
	}
}

SRMObj *SRMCtl::find(int seqnum)
{
	SRMObj *curr = m_list;

	while (curr != 0 && curr->m_seqnum != seqnum)
		curr = curr->m_next;

	return curr;
}

double SRMCtl::manage_timers(void)
{
	// get current sim time
	double curr_tm = Scheduler::instance().clock();

	// expire all the timers
	for (;;) {
		SRMObj *expired;

		if (m_list == 0 || (m_list->m_expire > curr_tm))
			break;

		expired = m_list;
		m_list = m_list->m_next;

		// printf("expire - %d\n",expired->m_seqnum);
			   
		expired->expire(this);
		// The delete will be done in the handling routines if necessary
	}
	// return next sim time that we have something to do
	if (m_list == 0)
		return -1.0;

	return (m_list->m_expire - curr_tm);
}

void SRMCtl::handle_data(int seqnum)
{
	if (m_expected_seqnum >= 0) {
		if (seqnum > m_expected_seqnum) {
			// schedule NACKs for all lost
			for (int count = m_expected_seqnum; count < seqnum; count++) {
				SRMReq *req;
				double delay;

				m_agent->request_sched_hook();
				
				// Here is the request delay
				delay = m_agent->request_tmo();

				req = new SRMReq(SRM_PENDING_REQUEST,count);
				if (req == 0) {
					fprintf(stderr,"Could not allocate SRMReq\n");
					exit(1);
				}
				sched(req,delay);
			}
		}
	}
	m_expected_seqnum = seqnum + 1;
	// Ignore if less than!
}

void SRMCtl::handle_nack(int seqnum)
{
	SRMObj *obj = find(seqnum);

	if (obj != 0) {
		obj->recv_nack(this);
	} else {
		// We don't know about it so check with expected to see
		// if we repair
		if (seqnum < m_expected_seqnum) {
			SRMRep *rep;
			double delay;

			m_agent->repair_sched_hook();
			
			// Here is the repair delay
			delay = m_agent->repair_tmo();
			
			rep = new SRMRep(SRM_PENDING_REPAIR,seqnum);
			if (rep == 0) {
				fprintf(stderr,"Could not allocate SRMRep\n");
				exit(1);
			}
			sched(rep,delay);
		}
	}
}

void SRMCtl::handle_repair(int seqnum)
{
	SRMObj *obj = find(seqnum);

	if (obj != 0) {
		obj->recv_repair(this);
	} else {
		// if we don't know about it then do nothing!
	}
}

