#include "agent.h"
#include "node.h"
#include "Tcl.h"
#include "packet.h"
#include "trace.h"
#include "random.h"
#include "srm.h"

static class SRMMatcher : public Matcher {
public:
	SRMMatcher() : Matcher("agent") {}
	TclObject* match(const char* id) {
		if (strcmp(id, "srm") == 0)
			return (new SRM_Agent());
		return (0);
	}
} matcher_srm;

SRM_Agent::SRM_Agent()
	: Agent(-1), m_scheduled(0), m_reqs_sched(0),
	  m_reqs_sent(0), m_reps_sched(0), m_reps_sent(0), m_pend_reqs(0),
	  m_pend_reps(0), m_on_req_proc(0), m_on_req_sent_proc(0),
	  m_on_rep_proc(0), m_on_rep_sent_proc(0), m_on_repaired_proc(0)
{
	m_ctl.agent(this);
	Tcl &tcl = Tcl::instance();

	link_time("ns_srm","request-delay",&m_req_delay,0);
	link_time("ns_srm","repair-delay",&m_rep_delay,0);
	link_time("ns_srm","ignore-delay",&m_ignore_delay,0);
	link_real("ns_srm","ignore-constant",&m_ignore_const,0);
	link_real("ns_srm","request-c1",&m_c1,0);
	link_real("ns_srm","request-c2",&m_c2,0);
	link_real("ns_srm","repair-d1",&m_d1,0);
	link_real("ns_srm","repair-d2",&m_d2,0);
	link_int("ns_srm","repair-size",&m_rep_sz,0);
	link_int("ns_srm","request-size",&m_req_sz,0);
	link_real("ns_srm","backoff",&m_backoff_const,0);

	// some read-only variables
	link_int(0,"requests-sched",&m_reqs_sched,1);
	link_int(0,"requests-sent",&m_reqs_sent,1);
	link_int(0,"repairs-sched",&m_reps_sched,1);
	link_int(0,"repairs-sent",&m_reps_sent,1);
	link_int(0,"pending-requests",&m_pend_reqs,1);
	link_int(0,"pending-repairs",&m_pend_reps,1);
}

double SRM_Agent::request_tmo(void)
{
	return Random::uniform(m_c1*m_req_delay,(m_c1+m_c2)*m_req_delay);
}

double SRM_Agent::backoff_tmo(double last)
{
	return (last*m_backoff_const);
}

double SRM_Agent::repair_tmo(void)
{
	return Random::uniform(m_d1*m_rep_delay,(m_d1+m_d2)*m_rep_delay);
}

double SRM_Agent::ignore_tmo(void)
{
	return m_ignore_const*m_ignore_delay;
}

void SRM_Agent::request_sched_hook(void)
{
	m_pend_reqs++;
	m_reqs_sched++;
	if (m_on_req_proc != 0)
		Tcl::instance().eval(m_on_req_proc);
}

void SRM_Agent::repair_sched_hook(void)
{
	m_pend_reps++;
	m_reps_sched++;
	if (m_on_rep_proc != 0)
		Tcl::instance().eval(m_on_rep_proc);
}

void SRM_Agent::repaired_hook(int req)
{
	// Called when a requested repair is received
	if (req) {
		m_pend_reqs--;
		if (m_on_repaired_proc != 0)
			Tcl::instance().eval(m_on_repaired_proc);
	} else {
		m_pend_reps--;
	}
}

void SRM_Agent::timeout(int)
{
	double interval = m_ctl.manage_timers();

	if (interval != -1.0) {
		m_scheduled = 1;
		sched(interval,1);
	} else {
		m_scheduled = 0;
	}
}

void SRM_Agent::recv(Packet* pkt)
{
	// drop anything that comes back to us that we send
	if (pkt->src_ == node_->addr() &&
		pkt->sport_ == sport_) {
		Packet::free(pkt);
		return;
	}

	if (m_scheduled)
		cancel(1);
	
	if (pkt->type_ == PT_SRM_NACK) {
		m_ctl.handle_nack(pkt->seqno_);
	} else if (pkt->type_ == PT_SRM_REP) {
		m_ctl.handle_repair(pkt->seqno_);
	} else {
		m_ctl.handle_data(pkt->seqno_);
	}

	double interval = m_ctl.manage_timers();
	if (interval != -1.0) {
		m_scheduled = 1;
		sched(interval,1);
	} else {
		m_scheduled = 0;
	}
	
	Packet::free(pkt);
}

void SRM_Agent::send_nack(int sn)
{
	Packet *p = allocpkt();

	m_reqs_sent++;
	if (m_on_req_sent_proc != 0)
		Tcl::instance().eval(m_on_req_sent_proc);
	
	p->type_ = PT_SRM_NACK;
	p->seqno_ = sn;
	p->size_ = m_req_sz;
	node_->transmit(p);
}

void SRM_Agent::send_repair(int sn)
{
	Packet *p = allocpkt();

	m_reps_sent++;
	if (m_on_rep_sent_proc != 0)
		Tcl::instance().eval(m_on_rep_sent_proc);
	
	p->type_ = PT_SRM_REP;
	p->seqno_ = sn;
	p->size_ = m_rep_sz;
	node_->transmit(p);
}

/*
 * $proc on-request-proc $handler
 * $proc on-request-sent-proc $handler
 * $proc on-repair-proc $handler
 * $proc on-repair-sent-proc $handler
 */
int SRM_Agent::command(int argc, const char*const* argv)
{
	if (argc == 2) {
		if (strcmp(argv[1], "clear") == 0) {
			// reset SRMCtl (clear timers and expected_seqn)
			m_ctl.reset();
			m_pend_reqs = m_pend_reps = 0;
			m_reqs_sched = m_reqs_sent = 0;
			m_reps_sched = m_reps_sent = 0;
			return (TCL_OK);
		}
	} else if (argc == 3) {		
		if (strcmp(argv[1], "on-request-proc") == 0) {
			const char* proc = argv[2];
			int n = strlen(proc) + 1;
			delete [] m_on_req_proc;
			m_on_req_proc = new char[n];
			strcpy(m_on_req_proc, proc);
			return (TCL_OK);
		} else if (strcmp(argv[1], "on-request-sent-proc") == 0) {
			const char* proc = argv[2];
			int n = strlen(proc) + 1;
			delete [] m_on_req_sent_proc;
			m_on_req_sent_proc = new char[n];
			strcpy(m_on_req_sent_proc, proc);
			return (TCL_OK);
		} else if (strcmp(argv[1], "on-repair-proc") == 0) {
			const char* proc = argv[2];
			int n = strlen(proc) + 1;
			delete [] m_on_rep_proc;
			m_on_rep_proc = new char[n];
			strcpy(m_on_rep_proc, proc);
			return (TCL_OK);
		} else if (strcmp(argv[1], "on-repair-sent-proc") == 0) {
			const char* proc = argv[2];
			int n = strlen(proc) + 1;
			delete [] m_on_rep_sent_proc;
			m_on_rep_sent_proc = new char[n];
			strcpy(m_on_rep_sent_proc, proc);
			return (TCL_OK);
		} else if (strcmp(argv[1], "on-repaired-proc") == 0) {
			const char* proc = argv[2];
			int n = strlen(proc) + 1;
			delete [] m_on_repaired_proc;
			m_on_repaired_proc = new char[n];
			strcpy(m_on_repaired_proc, proc);
			return (TCL_OK);
		}
	}
	return (Agent::command(argc, argv));
}
