/*
 * Copyright (c) 1990 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by the University of California, Lawrence Berkeley Laboratory,
 * Berkeley, CA.  The name of the University may not be used to
 * endorse or promote products derived from this software without
 * specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */
/*
 * PSC Changes Copyright (c) 1996 Pittsburgh Supercomputing Center,
 *                     Jeff Semke, Matt Mathis, and Jamshid Mahdavi 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the modifications to this
 * software were developed by the Pittsburgh Supercomputing Center.
 * The name of the Center may not be used to endorse or promote
 * products derived from this software without specific prior written
 * permission.  THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE
 * IMPLIED WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE. 
 *
 */

#include <stdio.h>
#include <sys/types.h>

#include "tcp.h"
#include "scoreboardrh.h"
#include "random.h"
#include "trace.h"

#define TRUE    1
#define FALSE   0
#define RECOVER_DUPACK  1
#define RECOVER_TIMEOUT 2
#define RECOVER_QUENCH  3

class FackRHTcpAgent : public TcpAgent {
 public:
	FackRHTcpAgent();
	virtual void recv(Packet *pkt);
        virtual void timeout(int tno);
	void process_ack (Packet* pkt);
	int maxsack (Packet* pkt);
	void send(int force, int reason, int maxburst);
 protected:
	int retran_data_;
	ScoreBoardRH scb_;
        int hithresh_;    /* Rate halving bounding parameter */
        int lothresh_;    /* Rate halving bounding parameter */
        unsigned int t_flags_;   /* tcp control block flags for fack */
#define TF_RECOVERY    0x0400     /* We are recovering from a lost segment */
#define TF_REPAIRED    0x0800     /* We have retransmitted something */
#define TF_WHOLD       0x1000     /* we are in the window hold state */
#define TF_TOGGLE      0x2000     /* divide by 2 toggle */
};

static class FackRHTcpMatcher : public Matcher {
public:
	FackRHTcpMatcher() : Matcher("agent") {}
	TclObject* match(const char* type) {
		if (strcmp(type, "tcp-fackrh") == 0)
			return (new FackRHTcpAgent());
		return (0);
	}
} matcher_fack;

FackRHTcpAgent::FackRHTcpAgent() : retran_data_(0)
{
      t_flags_ = 0;
}

/*
 * Process an ack.
 */
void 
FackRHTcpAgent::process_ack(Packet* pkt)
{
      int ms, fack;

      last_ack_ = pkt->seqno_;
      if (highest_ack_ < last_ack_)
	    highest_ack_ = last_ack_;
      if (seqno_ < last_ack_ + 1)
	    seqno_ = last_ack_ + 1;
      ms = maxsack(pkt);
      if (scb_.GetFack() < ms)
	    scb_.SetFack(ms);
      if (scb_.GetFack() < last_ack_)
	    scb_.SetFack(last_ack_);
      fack = scb_.GetFack();
      if (maxseq_ < fack)
	    maxseq_ = fack;
      
      newtimer(pkt);
      if (rtt_active_ && pkt->seqno_ >= rtt_seq_) {
	    rtt_active_ = 0;
	    t_backoff_ = 1;
      }
      /* with timestamp option */
      double tao = Scheduler::instance().clock() - pkt->bd_.tcp_.ts_;
      rtt_update(tao);
      /* update average window */
      awnd_ *= 1.0 - wnd_th_;
      awnd_ += wnd_th_ * cwnd_;
}

int 
FackRHTcpAgent::maxsack(Packet *pkt) 
{
	int maxsack=-1, sack_index;
	
	for (sack_index=0; sack_index < pkt->bd_.tcp_.sa_length_; sack_index++) {
		if (pkt->bd_.tcp_.sa_right_[sack_index] > maxsack)
			maxsack = pkt->bd_.tcp_.sa_right_[sack_index];
	}
	return (maxsack-1);
}

void 
FackRHTcpAgent::recv(Packet *pkt) 
{
      double win;
      int fack;
      int acked;          /* Number of packets acked by current ack */
      
      if (pkt->type_ != PT_ACK) {
	    fprintf(stderr,
		    "ns: configuration error: tcp received non-ack\n");
	    exit(1);
      }
      if (pkt->flags_ & PF_ECN)
	    quench(1);
  
      if (pkt->seqno_ > maxseq_)  {
	    goto drop_;
      }
  
      acked = (int) pkt->seqno_ - last_ack_;

      if (acked < 0) {
	    goto drop_;
      }

      process_ack(pkt);

      retran_data_ -= scb_.UpdateScoreBoard(highest_ack_, maxseq_, pkt);

      /* Check for new lossage at the receiver (either dupacks or SACK).
       * All unexpected events are captured ASAP, including simple mis-ordering
       * of packets.  Initialize the Rate-Halving bounding parameters.
       * These will be cleared soon if it turns out to have been simple
       * mis-ordering.
       */
  
      if (!(t_flags_ & TF_RECOVERY) 
	  && (last_ack_ != maxseq_) && 
	  (pkt->bd_.tcp_.sa_length_ || !acked)) {
	    win = cwnd_ / 2;
	    if (win < 2) win = 2;
	    hithresh_ = (int)win;
	    win /= 2;
	    if (win < 2) win = 2;
	    ssthresh_ = lothresh_ = (int)win;
	    
	    recover_ = maxseq_;
	    t_flags_ |= TF_RECOVERY;
	    t_flags_ &= ~(TF_REPAIRED | TF_WHOLD | TF_TOGGLE);

	    // retran_data_ -= scb_.UpdateScoreBoard(highest_ack_, maxseq_, pkt);
      }
  
      // Manage recovery and implement Rate-Halving

      if (t_flags_ & TF_RECOVERY) {

	    /* Calculate the new window during recovery.
	     * First verify that awnd is at least 1 mss smaller than cwnd
	     * (to account for the current ACK).  Otherwise cwnd might
	     * grow under pathological conditions.  E.g. if the network is
	     * replicating ACKs.
	     */
	    fack = scb_.GetFack();
	    awnd_ = maxseq_ - fack + retran_data_;
	    if (awnd_ + 1 <= cwnd_) {
		  cwnd_ = awnd_;          // reduce cwnd by 1 normally
		  if (t_flags_ & TF_WHOLD)
			cwnd_++;          // restore cwnd (hold state)
		  else if ((fack >= recover_) 
			   || (awnd_ <= lothresh_)) {
			t_flags_ |= TF_WHOLD;
			cwnd_++;          // restore cwnd (begin hold state)
		  }
		  else if ((t_flags_ ^= TF_TOGGLE) & TF_TOGGLE)
			cwnd_++;          // restore cwnd for every 2 acks
	    }

	    /* Cancel recovery if we have all data.  If anything was repaired,
	     * we make the /2 window adjustment mandatory.
	     */
	    if ((int)pkt->bd_.tcp_.sa_length_==0
		&& scb_.IsEmpty()
		&& (!(t_flags_ & TF_REPAIRED) 
		    || (fack >= recover_))) {
		  if ((t_flags_ & TF_REPAIRED)
		      && (cwnd_ > hithresh_))
			cwnd_ = hithresh_;
		  retran_data_ = 0;
		  t_flags_ &= ~(TF_RECOVERY|TF_REPAIRED|TF_WHOLD|TF_TOGGLE);
	    }
      }
      else if (acked) {
	    /* The normal, non-recovery case:
	     * open the window per slow start or congestion avoidance.
	     */
	    opencwnd();
      }

drop_:
      Packet::free(pkt);
      send(FALSE, REASON_NORMAL, maxburst_);
      if (trace_) 
	    plot();

}

void 
FackRHTcpAgent::timeout(int tno)
{
      if (tno == TCP_TIMER_RTX) {
	    if (highest_ack_ > last_ack_)
		  last_ack_ = highest_ack_;
	    retran_data_ = 0;
	    recover_cause_ = 2;
	    closecwnd(0);			
	    scb_.ClearScoreBoard();
	    reset_rtx_timer(1);
	    scb_.SetFack(last_ack_);
	    seqno_ = last_ack_ + 1;
	    send(0, REASON_TIMEOUT);
      } else {
	    TcpAgent::timeout(tno);
      }
}


void 
FackRHTcpAgent::send(int force, int reason, int maxburst)
{
      register int pktno, found, nextpktno, npacket = 0;
      int win = window();
      int xmit_seqno;
      int off, clen;
      int fack;


      fack = scb_.GetFack();

      while ( seqno_ <= fack + win - retran_data_) { 
	    xmit_seqno = -1;
	    // data fits in window

	    if (overhead_ == 0 || force) { // silly window avoidance
		  
		  if (t_flags_ & TF_RECOVERY) 
			xmit_seqno = scb_.GetNextRetran();

		  if (cwnd_ > seqno_ - fack + retran_data_ - 1) {
			if (xmit_seqno == -1) {
			      xmit_seqno = seqno_++;
			}
			else 
			      if (t_flags_ & TF_RECOVERY) {
				    scb_.MarkRetran(xmit_seqno, seqno_);
				    retran_data_++;
				    t_flags_ |= TF_REPAIRED;
			      }

			output(xmit_seqno, reason);
			if (seqno_ <= xmit_seqno) {
			      printf("Hit a strange case 2.\n");
			      seqno_ = xmit_seqno + 1;
			}
			npacket++;
		  }
	    } else if (!pending_[TCP_TIMER_DELSND]) {
		  /*
		   * Set a delayed send timeout.
		   */
		  sched(Random::uniform(overhead_), TCP_TIMER_DELSND);
		  return;
	    }
	    
	    if (maxburst && (npacket == maxburst))
		  break;
      }
}
