#
# Copyright (c) 1997 The Regents of the University of California.
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 1. Redistributions of source code must retain the above copyright
#    notice, this list of conditions and the following disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions and the following disclaimer in the
#    documentation and/or other materials provided with the distribution.
# 3. All advertising materials mentioning features or use of this software
#    must display the following acknowledgement:
# 	This product includes software developed by the Network Research
# 	Group at Lawrence Berkeley National Laboratory.
# 4. Neither the name of the University nor of the Laboratory may be used
#    to endorse or promote products derived from this software without
#    specific prior written permission.
# 
# THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
# OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
#
# @(#) $Header: test-suite-lossy.tcl,v 1.2 97/03/21 14:45:40 kfall Exp $ (LBL)
#
# This test suite exercises lossy links.
#
# Create a simple 2-node drop-tail topology:
# 


#             0.8Mb,100ms
#           r1 --------- k1
#
proc create_testnet { } {

	global r1 k1
	set r1 [ns node]
	set k1 [ns node]
	set L [ns_duplex $r1 $k1 800Kb 100ms drop-tail]
	[lindex $L 0] set queue-limit 20
	[lindex $L 1] set queue-limit 20
}

proc create_lossynet { losstype } {

	global r1 k1
	set r1 [ns node]
	set k1 [ns node]
	set linkbw 800Kb
	set linkdel 100ms

	# r1->k1 is lossy link
	set link0 [ns link $r1 $k1 $losstype]
	$link0 set bandwidth $linkbw
	$link0 set delay $linkdel

	# k1->r1 is drop-tail
	set link1 [ns link $k1 $r1 drop-tail]
	$link1 set bandwidth $linkbw
	$link1 set delay $linkdel
	$link1 set queue-limit 20

	return $link0	; # return the lossy link
}

proc finish file {

	set f [open temp.rands w]
	puts $f "TitleText: $file"
	puts $f "Device: Postscript"
	
	exec rm -f temp.p temp.d 
	exec touch temp.d temp.p
	#
	# split queue/drop events into two separate files.
	# we don't bother checking for the link we're interested in
	# since we know only such events are in our trace file
	#
	exec awk {
		{
			if (($1 == "+" || $1 == "-" ) && \
			    ($5 == "tcp" || $5 == "ack"))\
					print $2, $8 + ($11 % 90) * 0.01
		}
	} out.tr > temp.p
	exec awk {
		{
			if ($1 == "d")
				print $2, $8 + ($11 % 90) * 0.01
		}
	} out.tr > temp.d

	puts $f \"packets
	flush $f
	exec cat temp.p >@ $f
	flush $f
	# insert dummy data sets so we get X's for marks in data-set 4
	puts $f [format "\n\"skip-1\n0 1\n\n\"skip-2\n0 1\n\n"]

	puts $f \"drops
	flush $f
	#
	# Repeat the first line twice in the drops file because
	# often we have only one drop and xgraph won't print marks
	# for data sets with only one point.
	#
	exec head -1 temp.d >@ $f
	exec cat temp.d >@ $f
	close $f
	exec xgraph -bb -tk -nl -m -x time -y packet temp.rands &
	
	exit 0
}

#
# Arrange for tcp source stats to be dumped for $tcpSrc every
# $interval seconds of simulation time
#
proc tcpDump { tcpSrc interval } {
	proc dump { src interval } {
		ns at [expr [ns now] + $interval] "dump $src $interval"
		puts [ns now]/cwnd=[$src get cwnd]/ssthresh=[$src get ssthresh]/ack=[$src get ack]
	}
	ns at 0.0 "dump $tcpSrc $interval"
}

proc tcpDumpAll { tcpSrc interval label } {
	proc dump { src interval label } {
		ns at [expr [ns now] + $interval] "dump $src $interval $label"
		puts $label/time=[ns now]/cwnd=[$src get cwnd]/ssthresh=[$src get ssthresh]/ack=[$src get ack]/rtt=[$src get rtt]
	}
	puts $label:window=[$tcpSrc get window]/packet-size=[$tcpSrc get packet-size]/bug-fix=[$tcpSrc get bug-fix]
	ns at 0.0 "dump $tcpSrc $interval $label"
}

proc openTrace { stopTime testName } {
	exec rm -f out.tr temp.rands
	global r1 k1    
	set traceFile [open out.tr w]
	ns at $stopTime \
		"close $traceFile ; finish $testName"
	set T [ns trace]
	$T attach $traceFile
	return $T
}

# use small window here to avoid loss at the drop-tail queue
#
proc test_noloss {} {
	global r1 k1
	create_testnet
	set tcp1 [ns_create_connection tcp-reno $r1 tcp-sink $k1 0]
	$tcp1 set window 10
	set ftp1 [$tcp1 source ftp]
	ns at 0.0 "$ftp1 start"

	tcpDump $tcp1 1.0

	# trace only the bottleneck link
	[ns link $r1 $k1] trace [openTrace 5.0 test_no_loss]

	ns run
}

proc test_probloss {} {
	global r1 k1
	set lossylink [create_lossynet lossy-uniform]
	$lossylink set loss_prob 0.05	; # a somewhat lossy link
	set tcp1 [ns_create_connection tcp-reno $r1 tcp-sink $k1 0]
	$tcp1 set window 10
	set ftp1 [$tcp1 source ftp]
	ns at 0.0 "$ftp1 start"

	tcpDump $tcp1 1.0

	# trace only the bottleneck link
	[ns link $r1 $k1] trace [openTrace 5.0 test_fivepercent_loss]

	ns run
}

proc test_detloss {} {
	global r1 k1
	set lossylink [create_lossynet lossy-det]
	$lossylink filter "0.0 1.0 3"
	$lossylink filter "0.0 1.0 4"
	$lossylink filter "0.0 1.0 5"
	set tcp1 [ns_create_connection tcp-reno $r1 tcp-sink $k1 0]
	$tcp1 set window 10
	set ftp1 [$tcp1 source ftp]
	ns at 0.0 "$ftp1 start"

	tcpDump $tcp1 1.0

	# trace only the bottleneck link
	[ns link $r1 $k1] trace [openTrace 5.0 test_deterministic_loss]

	ns run
}

proc test_periodicburstloss {} {
	global r1 k1
	set lossylink [create_lossynet lossy-patt]
	$lossylink set loss_interval 10
	$lossylink set loss_burst 3
	set tcp1 [ns_create_connection tcp-reno $r1 tcp-sink $k1 0]
	$tcp1 set window 10
	set ftp1 [$tcp1 source ftp]
	ns at 0.0 "$ftp1 start"

	tcpDump $tcp1 1.0

	# trace only the bottleneck link
	[ns link $r1 $k1] trace [openTrace 5.0 test_periodic_burst_loss]

	ns run
}

proc test_probburstloss {} {
	global r1 k1
	set lossylink [create_lossynet lossy-patt]
	$lossylink set loss_prob 0.05
	$lossylink set loss_burst 3
	set tcp1 [ns_create_connection tcp-reno $r1 tcp-sink $k1 0]
	$tcp1 set window 10
	set ftp1 [$tcp1 source ftp]
	ns at 0.0 "$ftp1 start"

	tcpDump $tcp1 1.0

	# trace only the bottleneck link
	[ns link $r1 $k1] trace [openTrace 5.0 test_prob_burst_loss]

	ns run
}
if { $argc != 1 } {
	puts stderr {usage: ns test-suite-lossy.tcl [ noloss probloss detloss periodicburstloss probburstloss ... ]}
	exit 1
}
if { "[info procs test_$argv]" != "test_$argv" } {
	puts stderr "test-suite-lossy.tcl: no such test: $argv"
}
test_$argv
